unit t4a_trashcan;
{
  t4a_trashcan

  Unit that provides a cross platform function, MoveToTrash(), to move files to the system trashcan.

  Version 1.0, 2020-01-09 (C) Hans Luijten, www.tweaking4all.com

  You're free to use this code in any of your projects, commercial or not.
  If you decide to include this in a library then please mention the source:
  https://www.tweaking4all.com/software-development/lazarus-development/lazarus-crossplatform-trashcan/
  or
  https://www.tweaking4all.com/

  Usage:

  You can move an individual file or individual directory to the trashcan as such:

  MoveToTrash(Filename):boolean;
  MoveToTrash(Directoryname):boolean;

  Note: this function will always return TRUE under macOS, since the API doesn't return a TRUE/FALSE
        and would need a callback function, which would make this rather complicated.
        Testing if File or Directory exists after deletion will not work reliably, since the API call works async.
}

{$mode objfpc}{$H+} 
{$IFDEF DARWIN}
  {$modeswitch ObjectiveC1}
  {$linkframework CoreFoundation}
{$ENDIF}

interface

uses
  Classes, SysUtils

  {$IFDEF LINUX}
  , DateUtils, LCLProc, LazFileUtils, FileUtil
  {$ENDIF}

  {$IFDEF WINDOWS}
  , ShellApi
  {$ENDIF}

  {$IFDEF DARWIN}
  , MacOSAll, CocoaAll
  {$ENDIF}
  ;

  {$IFDEF WINDOWS}
  function MoveToTrash(FileOrDirName: UnicodeString): Boolean;
  {$ELSE}
  function MoveToTrash(FileOrDirName: AnsiString): Boolean;
  {$ENDIF}


implementation

{ Move a file to the Trash, Trashcan or Recycle Bin }

{$IFDEF DARWIN}
  function MoveToTrash(FileOrDirName:AnsiString): boolean;
  var
    aNSArray : NSMutableArray;
    aNSURL   : NSURL;
  begin
    aNSArray := NSMutableArray(NSMutableArray.array_).init; // How can I create a new array?
    aNSURL   := NSURL.fileURLWithPath(NSSTR(pchar(FileOrDirName)));
    aNSArray.addObject(aNSURL);
    NSWorkspace.sharedWorkspace.recycleURLs_completionHandler(aNSArray,nil);

    // Fails: Because at test time the file/directory may still exist, since recycleURLs_completionHandler
    //        does an async delete, teh following check fails:
    // Result := not(FileExists(FileOrDirName)) and not(DirectoryExists(FileOrDirName));

    Result := true;
  end;
{$ENDIF}

{$IFDEF WINDOWS}
  function MoveToTrash(FileOrDirName: UnicodeString): Boolean;
  var
    R: TSHFILEOPSTRUCTW;
  begin
    FillChar(R, SizeOf(R), 0);
    R.wFunc := FO_DELETE;
    R.pFrom := PWideChar(FileOrDirName + #0);
    R.fFlags := FOF_ALLOWUNDO or FOF_NOCONFIRMATION;
    Result := SHFileOperationW(@R) = 0;
  end;
{$ENDIF}

{$IFDEF LINUX}
  function MoveToTrash(FileOrDirName:AnsiString): boolean;
  var
    TrashCanBase  : string = '~/.local/share/Trash/';
    TrashCanFiles : string = '~/.local/share/Trash/files/';
    TrashCanInfo  : string = '~/.local/share/Trash/info/';
    TrashCanAlt   : string;

    InfoFile      : TStringList;

    isDirectory               : boolean;
    newFileOrDirName          : string;
    NamePart1, NamePart2      : string;

    FilesInDirectory          : TStringList;
    counter                   : integer;

    function EscapeURLString(s: string): string;
    var
      i: integer;
      source: PAnsiChar;
    begin
      Result := '';
      source := pansichar(s);
      for i := 1 to length(source) do
        if not (source[i - 1] in ['A'..'Z', 'a'..'z', '0'..'9', '-', '_', '~', '.', ':', '/']) then
          Result := Result + '%' + IntToHex(Ord(source[i - 1]), 2)
        else
          Result := Result + source[i - 1];
    end;

  begin
    if RightStr(FileOrDirName,1)=DirectorySeparator then  // remove directory separator if this is the last char
      FileOrDirName := LeftStr(FileOrDirName,Length(FileOrDirName)-1);

    // Are we moving a directory into the trashcan?
    isDirectory   := DirectoryExists(FileOrDirName);

    TrashCanFiles := ExpandFileName(TrashCanFiles);
    TrashCanInfo  := ExpandFileName(TrashCanInfo);
    TrashCanBase  := ExpandFileName(TrashCanBase);

    if not DirectoryExists(TrashCanBase) then  // We didn't find the standard dir, so now we are guessing alternatives
      begin
        TrashCanAlt := '';

        if DirectoryExists(ExpandFileName('~/Desktop/Trash/')) then
          TrashCanAlt:=ExpandFileName('~/Desktop/Trash/')
        else if DirectoryExists(ExpandFileName('~/trash/')) then
          TrashCanAlt:=ExpandFileName('~/trash/')
        else if DirectoryExists(ExpandFileName('~/Trash/')) then
          TrashCanAlt:=ExpandFileName('~/Trash/')
        else if DirectoryExists(ExpandFileName('~/.trash/')) then
          TrashCanAlt:=ExpandFileName('~/.trash/')
        else if DirectoryExists(ExpandFileName('~/.Trash/')) then
          TrashCanAlt:=ExpandFileName('~/.Trash/');

        if TrashCanAlt='' then  // we didn't find an alternative: create standard dirs
          begin
             ForceDirectories(TrashCanFiles);
             ForceDirectories(TrashCanInfo);
          end
        else                   // we did find an alternative, make sure we use it for files (and not for trashinfo)
          begin
            TrashCanFiles := TrashCanAlt;
            TrashCanInfo  := '';
          end;
      end;

    newFileOrDirName := ExtractFileName(FileOrDirName);
    NamePart1 := LazFileUtils.ExtractFileNameWithoutExt(newFileOrDirName);
    NamePart2 := ExtractFileExt(newFileOrDirName);
    counter   := 0;

    // if file already in Trashcan, we add a number (eg. example.doc, example.2.doc, expample.3.doc etc.)
    while (isDirectory      and DirectoryExists(TrashCanFiles+newFileOrDirName)) or
          (not(isDirectory) and FileExists(TrashCanFiles+newFileOrDirName)) do
      begin
        inc(Counter);
        newFileOrDirName := NamePart1 + BoolToStr(Counter>1,'.'+IntToStr(Counter),'') + NamePart2;
      end;

    // Create a trashinfo file, if the standard dir did exist!
    if TrashCanInfo<>'' then
      begin
        InfoFile := TStringList.Create;
        InfoFile.Text := '[Trash Info]'+LineEnding+
                         'Path='+EscapeURLString(FileOrDirName)+LineEnding+
                         'DeletionDate='+FormatDateTime('YYYY-MM-DD',Now)+'T'+FormatDateTime('hh:nn:ss',Now)+LineEnding; //'2020-01-09T17:05:41'+
        InfoFile.SaveToFile(TrashCanInfo+newFileOrDirName+'.trashinfo');
        InfoFile.Free;
      end;

    // make new filename now full path
    newFileOrDirName := TrashCanFiles+newFileOrDirName;

    // Move File or Directory - try rename first, if that fails try copying files, and if copying worked delete originals
    Result := RenameFile(FileOrDirName,newFileOrDirName);  // try moving file or dir

    if not Result then // if rename failed then try copy and delete (aka move)
      begin
        if isDirectory then
          begin
            FilesInDirectory := FindAllFiles(FileOrDirName, '*', true);

            for Counter:=0 to FilesInDirectory.Count-1 do  // Copy the file structure to the trashcan
              begin
                if ForceDirectories( ExtractFilePath( StringReplace(FilesInDirectory.Strings[Counter],FileOrDirName,newFileOrDirName,[] ) ) ) then
                  begin
                    if not CopyFile( FilesInDirectory.Strings[Counter], StringReplace(FilesInDirectory.Strings[Counter],FileOrDirName,newFileOrDirName,[] ) ) then
                      begin  // failed copying file - cleanup and bail
                        Result := false;

                        if DeleteDirectory(newFileOrDirName,true) then // delete what we have copied so far - things are going sideways
                          RemoveDir(newFileOrDirName);

                        if FileExists(TrashCanInfo+newFileOrDirName+'.trashinfo') then
                          DeleteFile(TrashCanInfo+newFileOrDirName+'.trashinfo');

                        Exit;
                      end;
                  end
                else // failed creating dir - cleanup and bail
                  begin
                    Result := false;

                    if DeleteDirectory(newFileOrDirName,true) then // delete what we have copied so far - things are going side ways
                      RemoveDir(newFileOrDirName);

                        if FileExists(TrashCanInfo+newFileOrDirName+'.trashinfo') then
                          DeleteFile(TrashCanInfo+newFileOrDirName+'.trashinfo');

                    Exit;
                  end;
              end;

            // all went well, remove old dir
            if DeleteDirectory(FileOrDirName,true) then
              Result := RemoveDir(FileOrDirName);

            FilesInDirectory.Free;
          end
        else
          begin
            if CopyFile(FileOrDirName,newFileOrDirName,true,false) then         // If rename fails: copy file
              Result := DeleteFile(FileOrDirName)                               // If copy succeeded: remove original
            else
              Result := false;                                                  // Copy/Delete failed
          end;
      end;
  end;
{$ENDIF}


end.

